class KerasCallback(object):
    def __init__(self):
        self.validation_data = None
        self.model = None

    def set_params(self, params):
        self.params = params

    def set_model(self, model):
        self.model = model

    def on_epoch_begin(self, epoch, logs=None):
        pass

    def on_epoch_end(self, epoch, logs=None):
        pass

    def on_batch_begin(self, batch, logs=None):
        pass

    def on_batch_end(self, batch, logs=None):
        pass

    def on_train_begin(self, logs=None):
        pass

    def on_train_end(self, logs=None):
        print('')
        print('[swiftace] Training metrics logged successfully.')



"""Git related utilities"""
import os
from subprocess import call


def is_git():
    """Check whether we are in a git repository"""
    return call(['git', 'rev-parse']) == 0


def git_branch():
    return os.popen('git branch | grep \* | cut -d \' \' -f2').read().strip()


def git_remote():
    """Get the remote URL"""
    return os.popen('git config --get remote.origin.url').read().replace(".git\n", "").strip()


def git_current_commit():
    """Get the current commit"""
    return os.popen('git rev-parse HEAD').read().strip()


def git_root():
    """Get the root of the git repository"""
    return os.popen('git rev-parse --show-toplevel').read().strip()


def git_commit(message):
    """Create a new commit"""
    return os.system('cd ' + git_root() + ' && git add . &&  git commit -a -m "' + message + '" && cd -')


def git_push():
    """Push the branch to origin"""
    return os.system("git push origin " + git_branch())


def git_commit_push(message):
    """Create a commit and push to origin"""
    if is_git():
        git_commit(message)
        git_push()
        return {
            'remote': git_remote(),
            'branch': git_branch(),
            'commit': git_commit(message)
        }



import hashlib
from tqdm import tqdm_notebook
from time import sleep
from swiftace.utils import validate_project_name, api, init_experiment, linkify
from swiftace.utils.checksum import get_hash
from swiftace.utils.anaconda import save_conda_env
from swiftace.utils.jupyter import upload_notebook, set_notebook_name, get_notebook_name
from swiftace.utils.git import git_commit_push

WEB_APP_URL = 'https://swiftai-217215.firebaseapp.com'
current_proj = None
current_run_id = None


def init(project_name):
    """Initialize a project and create a new session"""
    init_project(project_name)
    new_session()
    set_notebook_name()


def init_project(name):
    """Initialize the given project"""
    global current_proj
    if current_proj is None or current_proj != name:
        validate_project_name(name)
        resp = api.create_project(name)
        if resp:
            current_proj = resp
    print(
        f'[swiftace] Initlalized project: "{current_proj}"')


def new_session(force=False):
    """Start a new session (run), if one doesn't exist"""
    global current_proj
    global current_run_id
    if current_run_id is None or force:
        resp = init_experiment(current_proj)
        if resp is not None:
            current_run_id = resp
    print(f'[swiftace] Tracking experiment: "{current_run_id}"')


def get_run_id():
    if not current_proj or not current_run_id:
        raise Exception(
            "Please initialize a project using 'swiftace.init' before logging.")
    return current_run_id


def log_dataset(path, include_hash=True):
    """Record the dataset (file or directory)"""
    run_id = get_run_id()
    data = {'dataset': path}
    msg = f"[swiftace] Dataset logged: '{path}'"
    if include_hash:
        print('[swiftace] Computing checksum..')
        dataset_hash = get_hash(path)
        data['dataset_hash'] = dataset_hash
        msg += f"\n[swiftace] Checksum: {dataset_hash}"
    api.post_block(run_id, data, 'datasets')
    print(msg)


def log_hyperparams(data):
    """Record hyperparameters for the current run"""
    run_id = get_run_id()
    api.post_block(run_id, data, 'hyperparams')
    print('[swiftace] Hyperparameters logged.')


def log_metrics(data):
    """Record metrics for the current run"""
    run_id = get_run_id()
    api.post_block(run_id, data, 'metrics')
    print('[swiftace] Metrics logged.')


def upload_file(path):
    run_id = get_run_id()
    api.upload_file(run_id, path)
    print(f'[swiftace] File "{path}" uploaded.')


def get_session_url():
    """Get the tracking URL for the current session"""
    global current_proj
    global current_run_id
    if current_proj is not None and current_run_id is not None:
        return f"{WEB_APP_URL}/projects/{current_proj}/experiments/{current_run_id}"
    else:
        return ""


def get_project_url():
    """Get the tracking URL for the current project"""
    global current_proj
    if current_proj is not None:
        return f"{WEB_APP_URL}/projects/{current_proj}"
    else:
        return ""


def commit(message):
    """Create a commit for the current experiment"""
    run_id = get_run_id()

    print('[swiftace] Capturing environment..')
    env_hash = save_conda_env(run_id)

    print('[swiftace] Saving Jupyter notebook...')
    notebook_id = upload_notebook(run_id)

    print('[swiftace] Creating git commit..')
    git_meta = git_commit_push(message)

    print('[swiftace] Finalizing commit...')
    data = {'env': env_hash, 'code': notebook_id,
            'git': git_meta, 'message': message}
    res = api.post_commit(run_id, data)
    experiment_id = res['experimentId']

    print(f'[swiftace] Commit {experiment_id} successful.')
    return linkify("View in Dashboard", WEB_APP_URL)



"""Everything other than the top-level API"""
import re
from IPython.core.display import display, HTML
from swiftace.utils import api

NAME_HINT = "Project name can only contain alphabets(a-z), digits(0-9), underscores (_) and hyphens (-)."


class TrackingError(Exception):
    """Error class for exceptions related to experiment tracking"""
    pass


def validate_project_name(name):
    """Validate the project name"""
    if not name:
        raise TrackingError("Project name cannot be empty.")
    if not re.match('^[A-Za-z0-9-_]+$', name):
        raise TrackingError(f"Invalid project name :'{name}'.\n{NAME_HINT}")


def init_experiment(project_slug):
    """Initializes experiment"""
    if not project_slug:
        print(
            "[swiftace] Please initialize project first using: swiftace.init(\"<PROJECT-NAME>\")")
        return
    return api.create_experiment_run(project_slug)


def linkify(text, url=None):
    if url is None:
        url = text
    return display(HTML(f'<a href="{url}" target="_blank">{text}</a>'))



"""Anaconda related utilities"""
import os
import yaml
import logging
from swiftace.utils import api


class CondaError(Exception):
    """Error class for Anaconda-related exceptions"""
    pass


def get_conda_bin():
    """Get the path to the Anaconda binary"""
    conda_bin = os.popen('echo $CONDA_EXE').read().strip()
    if conda_bin == '':
        if os.popen('conda').read().strip() == '':
            raise CondaError(
                'Anaconda binary not found. Please make sure the "conda" command is in your system PATH or the environment variable $CONDA_EXE points to the anaconda binary')
        else:
            conda_bin = 'conda'
    logging.info(f'Anaconda binary: {conda_bin}')
    return conda_bin


def get_conda_env_name():
    """Get the name of the active conda environment"""
    env_name = os.popen('echo $CONDA_DEFAULT_ENV').read().strip()
    if env_name == '':
        env_name = 'base'
    logging.info(f'Anaconda environment: {env_name}')
    return env_name


def read_conda_env():
    """Read the anaconda environment into a YAML object"""
    command = f'{get_conda_bin()} env export -n {get_conda_env_name()}'
    env_str = os.popen(command).read()
    if env_str == '':
        raise CondaError(
            f'Failed to read Anaconda environment using command: "{command}""')
    return env_str


def save_conda_env(run_id):
    """Read and save the Anaconda environment"""
    env_str = read_conda_env()
    name = get_conda_env_name()
    return api.upload_env_file(run_id, env_str, name=name)


def parse_conda_env(stream):
    """Parse conda environment details"""
    try:
        yml_data = yaml.load(stream)
        return yml_data
    except yaml.YAMLError as e:
        raise CondaError('Failed to parse Anaconda environment') from e


def update_conda_env(fname, env_name):
    """Create or update a conda environment"""
    command = f'{get_conda_bin()} env update -n {env_name} -f {fname}'
    print(command)
    out = os.system(command)
    if out != 0:
        raise CondaError('Failed to update conda environment!')


def load_conda_env(env_id, name):
    """Download and update an anaconda environment from server"""
    data = api.download_conda_env(env_id)
    fname = f'{env_id}.yml'
    env_file = open(fname, 'wb')
    env_file.write(data)
    env_file.close()
    update_conda_env(fname, name)
